// lib3d.c - main file for 3D library
// by Jum Hig '99
// update: 31/10/1999
//
// Engine uses Allegro and libzb for zbuffering

// TODO: Filter insts / landscape polys to ftx - 20
// TODO: Correct clipping of GCOL (check newer Allegro versions)
// TODO: INtegr/fixed ray-caster


/* Structure:
	init_allegro()
	check_cpu() (to return cpu_mmx flag)
	set video mode
	create rendering buffer
	set up RGB map
	set projection viewport
	set up frustum
	set extern float scene_gap
	for every frame:
		scene_start()
		get_camera_matrix()
		for every instance:
			filter instance to camera space
			xlate insts about their own axes
			xlate insts to world space
			apply camera matrix (xform to camera space)
			clip to frustum ( clip3d_f() )
			scene_polygon3d_f()
		next instance
		scene_render()
		blit to screen
	next frame

*/

#include <stdio.h>
#include <allegro.h>
#include "zbpoly3d.h"
#include "lib3d.h"

BITMAP *E3D_buffer;
BITMAP *E3D_zbuffer;
MATRIX_f mfcamera;
MATRIX_f mfrotate;
MATRIX_f mfscale;
MATRIX_f mfxform;
MATRIX_f mfparent;
MATRIX_f mfsemifinal;
MATRIX_f mffinal;
int polysdrawn, polysprocessed;
char mystring[80];

// default scene
ThreeDeeScene E3D_scene;


// *******************************************************************
// E3D_Init			- Start up 3D engine
// *******************************************************************
int E3D_Init(int w, int h, int num_models, int num_textures, int num_insts) {
	// TODO: allocate memory and pointers for above, and
	//			make our buffers etc, and
	//		 check some stuff (ie: that allegro has been init'ed first
	//			etc


	int i;

	fprintf(stderr, "Initialising E3D:\n");
	fprintf(stderr, "Using %s\n", allegro_id);

	// check size of E3D buffer
	if((w > SCREEN_W) || (h > SCREEN_H)) {
		fprintf(stderr, "E3D buffer size too big - must be <= screen size!\n");
		fprintf(stderr, "Resizing to screen size...\n");
		w = SCREEN_W; h = SCREEN_H;
	}

	// check that allegro_init's been done
	if(os_type == OSTYPE_UNKNOWN) {
		fprintf(stderr, "Please call allegro_init before calling E3D_Init!\n");
		exit(1);
	}
	fprintf(stderr, "OS type %d\n", os_type);

/*	// check colour depth
	if(_colour_depth != 16) {
		fprintf(stderr, "Please set colour depth to 16 before calling E3D_Init!\n");
		exit(1);
	}
*/
	// check gfx mode set
	if(gfx_driver == NULL) {
		fprintf(stderr, "Please set gfx driver before calling E3D_Init!\n");
		exit(1);
	}

	// create render buffer
	E3D_buffer = create_bitmap(w, h);
   if(E3D_buffer == NULL) {
		fprintf(stderr, "Error: E3D_buffer init failed!\n");
		exit(1);							// 1 = some kind of init fail
	}
	// create zbuffer
	E3D_zbuffer = create_zbuffer(w, h);
   if(E3D_zbuffer == NULL) {
		fprintf(stderr, "Error: E3D_zbuffer init failed!\n");
		exit(1);							// 1 = some kind of init fail
	}

	// set projection viewport
	set_projection_viewport(0, 0, w, h);

	// set up default stuff
	E3D_SetBackColour(0.0, 0.0, 0.0);
   E3D_SetLightColour(0, 1.0, 1.0, 1.0);
	E3D_SetFogColour(0.5, 0.5, 0.5);
	E3D_SetFogDistance(0.0);			// fog off by default
	E3D_scene.renderdist = 500.0;		// default render cutoff distance
	E3D_scene.background = NULL;

	fprintf(stderr, "Init OK\n");
	return 0;
}


// *******************************************************************
// E3D_SetCameraPosn	- Place camera
// *******************************************************************
void E3D_SetCameraPosn(float x, float y, float z) {
	E3D_scene.camera.posn.x = x;
	E3D_scene.camera.posn.y = y;
	E3D_scene.camera.posn.z = z;
}

// *******************************************************************
// E3D_SetCameraRotn	- Rotate camera
// *******************************************************************
// NB: HACK - needs to be fixed !!!!
// find equations to convert HDR to dirn and up vectors
// - rotate camera unit dirn vector by pitch (x-axis), then heading (y-ax)
// - rotate camera unit up vector by roll, then pitch, then heading
// - may have to make up own rotation mx
// HOWTO: rotate unit camera dirn and up vectors
void E3D_SetCameraRotn(float heading, float pitch, float roll) {

	// get rot mx for rotating camera
	get_rotation_matrix_f(&mfrotate,
			pitch,
			heading,
			roll);

	// apply rotn mx to camera unit dirn vector
	apply_matrix_f(&mfrotate,
			0.0, 0.0, 1.0,
			&E3D_scene.camera.dirn.x,
			&E3D_scene.camera.dirn.y,
			&E3D_scene.camera.dirn.z);

	// apply rotn mx to camera unit up vector
	apply_matrix_f(&mfrotate,
			0.0, -1.0, 0.0,
			&E3D_scene.camera.up.x,
			&E3D_scene.camera.up.y,
			&E3D_scene.camera.up.z);
}

// *******************************************************************
// E3D_SetLight	- set omni light position
// *******************************************************************
void E3D_SetLightPosn(float x, float y, float z) {

}

// *******************************************************************
// E3D_SetLight	- set inf light direction
// *******************************************************************
void E3D_SetLightDirn(short n, float x, float y, float z) {
	// no error checking
	E3D_scene.light[n].dx = x;
	E3D_scene.light[n].dy = y;
	E3D_scene.light[n].dz = z;
}

// *******************************************************************
// E3D_SetLightColour	- set light colour
// *******************************************************************
void E3D_SetLightColour(short n, float r, float g, float b) {
	// no error checking
	E3D_scene.light[n].r = r;
	E3D_scene.light[n].g = g;
	E3D_scene.light[n].b = b;
}

// *******************************************************************
// E3D_SetBackColour	- set backgound colour
// *******************************************************************
void E3D_SetBackColour(float r, float g, float b) {
	// no error checking
	E3D_scene.backcolour.r = r;
	E3D_scene.backcolour.g = g;
	E3D_scene.backcolour.b = b;
}

// *******************************************************************
// E3D_SetFogColour - set fog colour   (0-255)
// *******************************************************************
void E3D_SetFogColour(float r, float g, float b) {

	// NB: no error checking (neccessary?)
	E3D_scene.fogcolour.r = r;
	E3D_scene.fogcolour.g = g;
	E3D_scene.fogcolour.b = b;
}

// *******************************************************************
// E3D_SetFogDistance - set fog wipeout distance
// *******************************************************************
void E3D_SetFogDistance(float dist) {

	// NB: no error checking (neccessary?)
	E3D_scene.fogdistance = dist;
	E3D_scene.renderdist = dist;
}

// *******************************************************************
// E3D_CreateInst	- add an instance to the scene
// *******************************************************************
ThreeDeeInstance *E3D_CreateInst(ThreeDeeModel *model) {
	int i;
	ThreeDeeScene *scene;

	scene = &E3D_scene;
	i = scene->numinsts;
	// create the instance
	scene->pinst[i] = (ThreeDeeInstance *)malloc(sizeof(ThreeDeeInstance));
	if(scene->pinst[i] == NULL) return NULL;
	// initialise instance parameters
	scene->pinst[i]->flags = 0;
	scene->pinst[i]->posn.x = 0.0;
	scene->pinst[i]->posn.y = 0.0;
	scene->pinst[i]->posn.z = 0.0;
	scene->pinst[i]->rotn.x = 0.0;
	scene->pinst[i]->rotn.y = 0.0;
	scene->pinst[i]->rotn.z = 0.0;
	scene->pinst[i]->scal.x = 1.0;
	scene->pinst[i]->scal.y = 1.0;
	scene->pinst[i]->scal.z = 1.0;
	// set pointer to model for this inst
	scene->pinst[i]->model = model;
	// set attach to NULL
	scene->pinst[i]->attach = NULL;
	scene->numinsts++;

	return scene->pinst[i];
}

// *******************************************************************
// E3D_RotateInst	- Set instance rotation
// *******************************************************************
void E3D_RotateInst(ThreeDeeInstance *pinst, float rx, float ry, float rz) {
	pinst->rotn.x = rx;
	pinst->rotn.y = ry;
	pinst->rotn.z = rz;
}

// *******************************************************************
// E3D_ScaleInst	- Set instance scale
// *******************************************************************
void E3D_ScaleInst(ThreeDeeInstance *pinst, float sx, float sy, float sz) {
	pinst->scal.x = sx;
	pinst->scal.y = sy;
	pinst->scal.z = sz;
}

// *******************************************************************
// E3D_XlateInst	- move instance
// *******************************************************************
void E3D_XlateInst(ThreeDeeInstance *pinst, float x, float y, float z) {
	pinst->posn.x = x;
	pinst->posn.y = y;
	pinst->posn.z = z;
}

// *******************************************************************
// E3D_DestroyInst	- Free up mem of inst n (NEEDED?)
// *******************************************************************
int E3D_DestroyInst(ThreeDeeInstance *pinst) {
	free(pinst);
	return 0;
}

// *******************************************************************
// E3D_DrawInst	- draw an instance (DEBUG!)
// *******************************************************************
void E3D_DrawInst(ThreeDeeInstance *pinst) {

	int j, k, t2;
	float ftx, fty, ftz;
	V3D_f vtemp[50];				// allegro V3D for compatibility later
	V3D_f vtemp2[12];
	V3D_f	vtemp3[12];
	V3D_f	vtemp4[12];
	rgb_f vcol[12];				// temp vertex colours for calcs
	rgb_f le;						// total light effect (temp calc)
	V3D_f *v_clip_in[12];
	V3D_f *v_clip_out[12];
	V3D_f *v_clip_tmp[12];
	V3D_f tnorm;
	int v_clip_outflags[12];
	int numvout;
	V3D_f *vout[12];
	ThreeDeeModel *thismodel;
	ThreeDeeScene *scene;
	ThreeDeeInstance *parent;
	float r, g, b, dr, dg, db, z, theta;
	float fog_ratio[12];
	int fogcol;

	// initialise pntrs to scene struct and this insts model
	thismodel = pinst->model;
	scene = &E3D_scene;
	parent = NULL;
	fogcol = makecol(scene->fogcolour.r * 255, scene->fogcolour.g *255, scene->fogcolour.b * 255);

	// initialise clipping pointers
	for(k=0; k<12; k++) {
		v_clip_out[k] = &vtemp3[k];
		v_clip_tmp[k] = &vtemp4[k];
	}

	// DEBUG - draw top view of inst in WORLD SPACE
	line(E3D_buffer, 480, 0, 480, 479, makecol(0, 128, 0));
	line(E3D_buffer, 320, 120, 639, 120, makecol(0, 128, 0));
	line(E3D_buffer, 0, 360, 639, 360, makecol(0, 128, 0));
	textout_centre(E3D_buffer, font, "World Space", 480, 2, makecol(0,255,255));

	for(j=0; j < thismodel->numverts; j++) {
		putpixel(E3D_buffer, 480 + (int) (pinst->posn.x + thismodel->vert[j].x),
						120 - (int) (pinst->posn.z - thismodel->vert[j].z),
						makecol(255, 255, 0));
	}
	circle(E3D_buffer, 480 + (int) (scene->camera.posn.x),
					120 - (int) (scene->camera.posn.z), 3, makecol(255, 128, 0));
	line(E3D_buffer, 480 + (int) (scene->camera.posn.x),
					120 - (int) (scene->camera.posn.z),
					480 + (int) (scene->camera.posn.x + scene->camera.dirn.x*20.0),
					120 - (int) (scene->camera.posn.z + scene->camera.dirn.z*20.0),
					makecol(0, 255, 255));



	// DEBUG - draw top view of scene in CAMERA SPACE
	textout_centre(E3D_buffer, font, "Camera Space", 480, 242, makecol(0,255,255));
	circle(E3D_buffer, 480, 360, 3, makecol(0,255,255));
	line(E3D_buffer, 480, 360, 480, 341, makecol(128,128,128));


	// filter this inst to the view frustum
	// (if attached, use parent inst's values)
	if(pinst->attach == NULL) {
		apply_matrix_f(&mfcamera,
				pinst->posn.x,
				pinst->posn.y,
				pinst->posn.z,
				&ftx, &fty, &ftz);
	}
	else {		// attached to another (parent) inst
		parent = (ThreeDeeInstance *)pinst->attach;
		apply_matrix_f(&mfcamera,
				parent->posn.x,
				parent->posn.y,
				parent->posn.z,
				&ftx, &fty, &ftz);
	}
//		fprintf(stderr, "inst %d  ftx %3.3f  fty %3.3f  ftz %3.3f\n", i, ftx, fty, ftz);

	// show inst in camera space before filtering
	putpixel(E3D_buffer, 480 + (int) (ftx),
			360 - (int) (ftz), makecol(255,128,0));

	// NB: distance cut-off done in Draw_Scene()!
	// do zero cutoff
	if(ftz < -20.0) return 1;

	// then filter
	if(ftz > (ftx - 20.0)) {
		if(ftz > (-ftx -20.0)) {
			// todo: filter to z = y and z = -y planes
			// instance is in view frustum
			putpixel(E3D_buffer, 480 + (int) (ftx),
					360 - (int) (ftz), makecol(0,255,0));
			// make xform in model space
			get_transformation_matrix_f(&mfxform, pinst->scal.x,
					pinst->rotn.x,
					pinst->rotn.y,
					pinst->rotn.z,
					pinst->posn.x,
					pinst->posn.y,
					pinst->posn.z);

			// if attached, xform relative to parent
			if(parent != NULL) {
				// make parent xform in world space
				get_transformation_matrix_f(&mfparent, parent->scal.x,
						parent->rotn.x,
						parent->rotn.y,
						parent->rotn.z,
						parent->posn.x,
						parent->posn.y,
						parent->posn.z);
				// multiply with inst's modelspace xform
				matrix_mul_f(&mfxform, &mfparent, &mfsemifinal);
				// multiply with camera xform
				matrix_mul_f(&mfsemifinal, &mfcamera, &mffinal);

				// get rot mx for rotating normals
				get_rotation_matrix_f(&mfrotate,
						pinst->rotn.x + parent->rotn.x,
						pinst->rotn.y + parent->rotn.y,
						pinst->rotn.z + parent->rotn.z);
			}
			else {	// not attached
				// combine with xform to camera space
				matrix_mul_f(&mfxform, &mfcamera, &mffinal);

				// get rot mx for rotating normals
				get_rotation_matrix_f(&mfrotate,
						pinst->rotn.x,
						pinst->rotn.y,
						pinst->rotn.z);
			}

			// apply our matrix to all verts in this inst's model
//			t1 = pinst->model;
			for(j=0; j < thismodel->numverts; j++) {
				apply_matrix_f(&mffinal,
						thismodel->vert[j].x,
						thismodel->vert[j].y,
						thismodel->vert[j].z,
						&vtemp[j].x, &vtemp[j].y, &vtemp[j].z);
//				putpixel(E3D_buffer, 480 + (int) (vtemp[j].x),
//						360 - (int) (vtemp[j].z), makecol(255,255,128));
			}	// next j (next vert)



// *** NB: 3D clipping must be before persp projection;
//			ie: will have to generate polygon, then clip, then persp proj
//				it's vertices (ie: big ball ache)
// *** NB2: can optimise by backface culling before clipping or
//				projecting.

			// *** GENERATE THE POLYGONS FOR THIS INSTANCE ***
			for(j=0; j < thismodel->numpolys; j++) {
				polysprocessed++;
//fprintf(stderr, "model %d numpolys %d\n", t1, thismodel->numpolys);
				// copy verts to polytemp and
				// create list of pointers to verts in this poly
				// 2 versions - 1 for fog, 1 for no fog
				// NOTE: can optimise by moving fog calcs to after
				// 		backface?
				// precalc fog effects for this poly's vertices
				if(scene->fogdistance > 0.0) {
					for(k=0; k < thismodel->poly[j].numverts; k++) {
						t2 = thismodel->poly[j].p[k];
						fog_ratio[k] = vtemp[t2].z / scene->fogdistance;
						if(fog_ratio[k] > 1.0) fog_ratio[k] = 1.0;
					}
				}
				//
				for(k=0; k < thismodel->poly[j].numverts; k++) {
					t2 = thismodel->poly[j].p[k];
					// optimise colour resolution by pre-converting rgb's
					// to c depending on colour mode
					// ie: before main game loop, *NOT* here!
					switch(thismodel->poly[j].filling) {
						case 0 :	// solid colour (flat "shaded")
							// only need colour for first vert
							if(k == 0) {
								vcol[0].r = thismodel->poly[j].colour.r;
								vcol[0].g = thismodel->poly[j].colour.g;
								vcol[0].b = thismodel->poly[j].colour.b;
							}
							break;
						case 1 :	// colour gradient
							vcol[k].r = thismodel->poly[j].vcol[k].r;
							vcol[k].g = thismodel->poly[j].vcol[k].g;
							vcol[k].b = thismodel->poly[j].vcol[k].b;
							break;
						case 2 :	// texture-mapped
							vtemp[t2].c = makecol((int) (thismodel->poly[j].vcol[k].r*255),
									(int) (thismodel->poly[j].vcol[k].g*255),
									(int) (thismodel->poly[j].vcol[k].b*255));
							vtemp[t2].u = thismodel->poly[j].texp[k].x;
							vtemp[t2].v = thismodel->poly[j].texp[k].y;
							break;
					}	// end switch

					v_clip_in[k] = &vtemp[t2];
// debug
				}	// next k (next vert)

				// *** CLIP POLYGON TO FRUSTUM ***
				//int clip3d_f(int type, float min_z, float max_z,
				//		int vc, V3D_f *vin[], V3D_f *vout[],
				//		V3D_f *vtmp[], int outflags[]);
				numvout = clip3d_f(POLYTYPE_FLAT, 1.0, E3D_scene.renderdist,
									thismodel->poly[j].numverts,
									v_clip_in, v_clip_out,
									v_clip_tmp, v_clip_outflags);

//numvout = thismodel->poly[j].numverts;

				// perpective projection
				for(k=0; k < numvout; k++) {
					persp_project_f(v_clip_out[k]->x, -v_clip_out[k]->y,
							v_clip_out[k]->z,
							&vtemp2[k].x, &vtemp2[k].y);
					vtemp2[k].z = v_clip_out[k]->z;
					vtemp2[k].c = v_clip_out[k]->c;
					vtemp2[k].u = v_clip_out[k]->u;
					vtemp2[k].v = v_clip_out[k]->v;
					vout[k] = &vtemp2[k];
//					putpixel(E3D_buffer, (int) (vtemp2[k].x),
//							240 - (int) (vtemp2[k].y), makecol(255,128,0));
// debug
//fprintf(stderr, "inst %d  poly %d  %3.3f %3.3f %3.3f %d\n",
//				i, j, vout[k]->x, vout[k]->y, vout[k]->z, vout[k]->c);

				}	// next k

				// *** BACKFACE CULL ***
				// (backface cull on persp proj poly!)
				if(polygon_z_normal_f(vout[0], vout[1], vout[2]) < 0.0) {
					// draw zbuffered polygon
//					polygon3d_f(E3D_buffer, POLYTYPE_FLAT,
//							scene->ptex[0], numvout, vout);
					// *** LIGHT SOURCE ***
					// NB: fogging must take place AFTER light-sourcing!
					// (to look correct)
					// This means we must store "fog %" for every vertex.
					apply_matrix_f(&mfrotate,
						thismodel->normal[j].x,
						thismodel->normal[j].y,
						thismodel->normal[j].z,
               	&tnorm.x, &tnorm.y, &tnorm.z);
// a.b = |a||b|cos(theta)
					theta = dot_product_f(tnorm.x, tnorm.y, tnorm.z,
										scene->light[0].dx,
										scene->light[0].dy,
										scene->light[0].dz);
// poly brightness = poly colour * (ambient "colour factors" + light sourcing)
					// for polys facing away from the light, no effect
					if(theta < 0.0) theta = 0.0;

					// calc effect of ambient + distant lights
					le.r = scene->ambient.r + theta * scene->light[0].r;
					le.g = scene->ambient.g + theta * scene->light[0].g;
					le.b = scene->ambient.b + theta * scene->light[0].b;
               if(le.r > 1.0) le.r = 1.0;
					if(le.g > 1.0) le.g = 1.0;
					if(le.b > 1.0) le.b = 1.0;


					// *** RENDER ***
					polysdrawn++;
					// what type of fill does it have?
					switch(thismodel->poly[j].filling) {
						case 0 :		// flat 	(solid SHADE)
							// calc colour for vert 0 only
							// NB: OPTIMISE!
							r = vcol[0].r * le.r;
							g = vcol[0].g * le.g;
							b = vcol[0].b * le.b;
							dr = r - scene->fogcolour.r;
							dg = g - scene->fogcolour.g;
							db = b - scene->fogcolour.b;
							vcol[0].r = r - dr * fog_ratio[0];
							vcol[0].g = g - dg * fog_ratio[0];
							vcol[0].b = b - db * fog_ratio[0];

							vout[0]->c = makecol(vcol[0].r * 255,
										vcol[0].g * 255,
										vcol[0].b * 255);
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_FLAT,
									scene->ptex[0], numvout, vout);
							break;
						case 1 :		// gouraud
							// calc colour for all verts
							for(k=0; k < numvout; k++) {
								r = vcol[k].r * le.r;
								g = vcol[k].g * le.g;
								b = vcol[k].b * le.b;
								dr = r - scene->fogcolour.r;
								dg = g - scene->fogcolour.g;
								db = b - scene->fogcolour.b;
								vcol[k].r = r - dr * fog_ratio[k];
								vcol[k].g = g - dg * fog_ratio[k];
								vcol[k].b = b - db * fog_ratio[k];
								vout[k]->c = makecol(vcol[k].r * 255,
										vcol[k].g * 255,
										vcol[k].b * 255);
							}
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_GCOL,
									scene->ptex[0], numvout, vout);
							break;
						case 2 :		// textured
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_PTEX,
									scene->ptex[0], numvout, vout);
							break;
					}

// DEBUG ***
//blit(E3D_buffer, screen, 0, 0, 0, 0, 640, 480);
//rest(500);
//return 0;

					// DEBUG - display vert values
					if(key[KEY_V]) {
//						for(k=0; k<scene->pmodel[t1].poly[j].numverts; k++) {
//							fprintf(stderr, "Vert %d: addr %d   x %3.3f   y %3.3f   z %3.3f\n",
//								k, vout[k], vout[k]->x, vout[k]->y, vout[k]->z);
//						}
					}
//						fprintf(stderr, "Polys drawn: %d \n", polysdrawn);

				}	// end if polygon_z_normal > 0.0
			}	// next j (next poly)
		}	 // end if 2
	}	// end if 1

}

// *******************************************************************
// E3D_DrawLandscape	- draw a landscape object
// *******************************************************************
void E3D_DrawLandscape(void) {

	int i, j, k, t2;
	float ftx, fty, ftz;
	V3D_f vtemp[50];				// allegro V3D for compatibility later
	V3D_f vtemp2[12];
	V3D_f	vtemp3[12];
	V3D_f	vtemp4[12];
	rgb_f vcol[12];				// temp vertex colours for calcs
	rgb_f le;						// total light effect (temp calc)
	V3D_f *v_clip_in[12];
	V3D_f *v_clip_out[12];
	V3D_f *v_clip_tmp[12];
	V3D_f tnorm;
	int v_clip_outflags[12];
	int numvout;
	V3D_f *vout[12];
	ThreeDeeLandscape *landmodel;
	ThreeDeeInstance *pland;
	ThreeDeeScene *scene;
	float r, g, b, dr, dg, db, z, theta;
	float fog_ratio[12];
	int fogcol;

	// initialise pntrs to scene struct and this
	landmodel = E3D_scene.plandmodel;
	pland = E3D_scene.pland;
	scene = &E3D_scene;
	fogcol = makecol(scene->fogcolour.r * 255, scene->fogcolour.g *255, scene->fogcolour.b * 255);

	// initialise clipping pointers
	for(k=0; k<12; k++) {
		v_clip_out[k] = &vtemp3[k];
		v_clip_tmp[k] = &vtemp4[k];
	}

	// DEBUG - draw top view of landscape in WORLD SPACE
/*	line(E3D_buffer, 480, 0, 480, 479, makecol(0, 128, 0));
	line(E3D_buffer, 320, 120, 639, 120, makecol(0, 128, 0));
	line(E3D_buffer, 0, 360, 639, 360, makecol(0, 128, 0));
	textout_centre(E3D_buffer, font, "World Space", 480, 2, makecol(0,255,255));

	for(j=0; j < thismodel->numverts; j++) {
		putpixel(E3D_buffer, 480 + (int) (pinst->posn.x + thismodel->vert[j].x),
						120 - (int) (pinst->posn.z - thismodel->vert[j].z),
						makecol(255, 255, 0));
	}
	circle(E3D_buffer, 480 + (int) (scene->camera.posn.x),
					120 - (int) (scene->camera.posn.z), 3, makecol(255, 128, 0));
	line(E3D_buffer, 480 + (int) (scene->camera.posn.x),
					120 - (int) (scene->camera.posn.z),
					480 + (int) (scene->camera.posn.x + scene->camera.dirn.x*20.0),
					120 - (int) (scene->camera.posn.z + scene->camera.dirn.z*20.0),
					makecol(0, 255, 255));



	// DEBUG - draw top view of scene in CAMERA SPACE
	textout_centre(E3D_buffer, font, "Camera Space", 480, 242, makecol(0,255,255));
	circle(E3D_buffer, 480, 360, 3, makecol(0,255,255));
	line(E3D_buffer, 480, 360, 480, 341, makecol(128,128,128));
*/

	// First, generate the xform matrices
	// make xform in model space
	get_transformation_matrix_f(&mfxform, pland->scal.x,
				pland->rotn.x,
				pland->rotn.y,
				pland->rotn.z,
				pland->posn.x,
				pland->posn.y,
				pland->posn.z);
	// combine with xform to camera space
	matrix_mul_f(&mfxform, &mfcamera, &mffinal);
	// get rot mx for rotating normals
	get_rotation_matrix_f(&mfrotate,
				pland->rotn.x,
				pland->rotn.y,
				pland->rotn.z);


	// for every poly in the landscape
	for(j=0; j<landmodel->numpolys; j++) {
		polysprocessed++;
		// filter this landscape poly to the view frustum
		apply_matrix_f(&mffinal,
				landmodel->centroid[j].x,
				landmodel->centroid[j].y,
				landmodel->centroid[j].z,
				&ftx, &fty, &ftz);
//		fprintf(stderr, "inst %d  ftx %3.3f  fty %3.3f  ftz %3.3f\n", i, ftx, fty, ftz);

	// show inst in camera space before filtering
//	putpixel(E3D_buffer, 480 + (int) (ftx),
//			360 - (int) (ftz), makecol(255,128,0));

		// NB: do distance filter!
		if((ftz < E3D_scene.renderdist) && (ftz > 0.0)) {

		// then filter
		if(ftz > ftx) {
			if(ftz > -ftx) {
				// todo: filter to z = y and z = -y planes
				// instance is in view frustum
//			putpixel(E3D_buffer, 480 + (int) (ftx),
//					360 - (int) (ftz), makecol(0,255,0));
//			t1 = pinst->model;
//			thismodel = scene->pmodel[t1];
				// xform this polys verts into temp verts
				for(k=0; k < landmodel->poly[j].numverts; k++) {
					t2 = landmodel->poly[j].p[k];
					apply_matrix_f(&mffinal,
							landmodel->vert[t2].x,
							landmodel->vert[t2].y,
							landmodel->vert[t2].z,
							&vtemp[k].x, &vtemp[k].y, &vtemp[k].z);
					// precalc fog effects for this poly's vertices
					if(scene->fogdistance > 0.0) {
						fog_ratio[k] = vtemp[k].z / scene->fogdistance;
						if(fog_ratio[k] > 1.0) fog_ratio[k] = 1.0;
					}
					// optimise colour resolution by pre-converting rgb's
					// to c depending on colour mode
					// ie: before main game loop, *NOT* here!
					switch(landmodel->poly[j].filling) {
						case 0 :	// solid colour (flat "shaded")
							// only need colour for first vert
							if(k == 0) {
								vcol[0].r = landmodel->poly[j].colour.r;
								vcol[0].g = landmodel->poly[j].colour.g;
								vcol[0].b = landmodel->poly[j].colour.b;
							}
							break;
						case 1 :	// colour gradient
							vcol[k].r = landmodel->poly[j].vcol[k].r;
							vcol[k].g = landmodel->poly[j].vcol[k].g;
							vcol[k].b = landmodel->poly[j].vcol[k].b;
							break;
						case 2 :	// texture-mapped
							vtemp[k].c = makecol((int) (landmodel->poly[j].vcol[k].r*255),
									(int) (landmodel->poly[j].vcol[k].g*255),
									(int) (landmodel->poly[j].vcol[k].b*255));
							vtemp[k].u = landmodel->poly[j].texp[k].x;
							vtemp[k].v = landmodel->poly[j].texp[k].y;
							break;
					}	// end switch
//				putpixel(E3D_buffer, 480 + (int) (vtemp[j].x),
//						360 - (int) (vtemp[j].z), makecol(255,255,128));

// *** NB: 3D clipping must be before persp projection;
//			ie: will have to generate polygon, then clip, then persp proj
//				it's vertices (ie: big ball ache)
// *** NB2: can optimise by backface culling before clipping or
//				projecting.

					v_clip_in[k] = &vtemp[k];
// debug
				}	// next k (next vert)

				// *** CLIP POLYGON TO FRUSTUM ***
				//int clip3d_f(int type, float min_z, float max_z,
				//		int vc, V3D_f *vin[], V3D_f *vout[],
				//		V3D_f *vtmp[], int outflags[]);
				// what type of fill does it have?
				t2 = POLYTYPE_FLAT;
				switch(landmodel->poly[j].filling) {
					case 0 :	// flat
						t2 = POLYTYPE_FLAT; break;
					case 1 :	// gradient fill
						t2 = POLYTYPE_GCOL; break;
					case 2 :	// tex fill
						t2 = POLYTYPE_PTEX; break;
				}
				numvout = clip3d_f(t2, 1.0, E3D_scene.renderdist,
									landmodel->poly[j].numverts,
									v_clip_in, v_clip_out,
									v_clip_tmp, v_clip_outflags);

//numvout = landmodel->poly[j].numverts;

				// perpective projection
				for(k=0; k < numvout; k++) {
					persp_project_f(v_clip_out[k]->x, -v_clip_out[k]->y,
							v_clip_out[k]->z,
							&vtemp2[k].x, &vtemp2[k].y);
					vtemp2[k].z = v_clip_out[k]->z;
					vtemp2[k].c = v_clip_out[k]->c;
					vtemp2[k].u = v_clip_out[k]->u;
					vtemp2[k].v = v_clip_out[k]->v;
					vout[k] = &vtemp2[k];
//					putpixel(E3D_buffer, (int) (vtemp2[k].x),
//							240 - (int) (vtemp2[k].y), makecol(255,128,0));
// debug
//fprintf(stderr, "inst %d  poly %d  %3.3f %3.3f %3.3f %d\n",
//				i, j, vout[k]->x, vout[k]->y, vout[k]->z, vout[k]->c);

				}	// next k

				// *** BACKFACE CULL ***
				// (backface cull on persp proj poly!)
				if(polygon_z_normal_f(vout[0], vout[1], vout[2]) < 0.0) {
					// draw zbuffered polygon
//					polygon3d_f(E3D_buffer, POLYTYPE_FLAT,
//							scene->ptex[0], numvout, vout);
					// *** LIGHT SOURCE ***
					// NB: fogging must take place AFTER light-sourcing!
					// (to look correct)
					// This means we must store "fog %" for every vertex.
					apply_matrix_f(&mfrotate,
						landmodel->normal[j].x,
						landmodel->normal[j].y,
						landmodel->normal[j].z,
               	&tnorm.x, &tnorm.y, &tnorm.z);
// a.b = |a||b|cos(theta)
					theta = dot_product_f(tnorm.x, tnorm.y, tnorm.z,
										scene->light[0].dx,
										scene->light[0].dy,
										scene->light[0].dz);
// poly brightness = poly colour * (ambient "colour factors" + light sourcing)
					// for polys facing away from the light, no effect
					if(theta < 0.0) theta = 0.0;

					// calc effect of ambient + distant lights
					le.r = scene->ambient.r + theta * scene->light[0].r;
					le.g = scene->ambient.g + theta * scene->light[0].g;
					le.b = scene->ambient.b + theta * scene->light[0].b;
               if(le.r > 1.0) le.r = 1.0;
					if(le.g > 1.0) le.g = 1.0;
					if(le.b > 1.0) le.b = 1.0;


					// *** RENDER ***
					polysdrawn++;
					// what type of fill does it have?
					switch(landmodel->poly[j].filling) {
						case 0 :		// flat 	(solid SHADE)
							// calc colour for vert 0 only
							// NB: OPTIMISE!
							r = vcol[0].r * le.r;
							g = vcol[0].g * le.g;
							b = vcol[0].b * le.b;
							dr = r - scene->fogcolour.r;
							dg = g - scene->fogcolour.g;
							db = b - scene->fogcolour.b;
							vcol[0].r = r - dr * fog_ratio[0];
							vcol[0].g = g - dg * fog_ratio[0];
							vcol[0].b = b - db * fog_ratio[0];

							vout[0]->c = makecol(vcol[0].r * 255,
										vcol[0].g * 255,
										vcol[0].b * 255);
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_FLAT,
									scene->ptex[0], numvout, vout);
							break;
						case 1 :		// gouraud
							// calc colour for all verts
							for(k=0; k < numvout; k++) {
								r = vcol[k].r * le.r;
								g = vcol[k].g * le.g;
								b = vcol[k].b * le.b;
								dr = r - scene->fogcolour.r;
								dg = g - scene->fogcolour.g;
								db = b - scene->fogcolour.b;
								vcol[k].r = r - dr * fog_ratio[k];
								vcol[k].g = g - dg * fog_ratio[k];
								vcol[k].b = b - db * fog_ratio[k];
								vout[k]->c = makecol(vcol[k].r * 255,
										vcol[k].g * 255,
										vcol[k].b * 255);
							}
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_GCOL,
									scene->ptex[0], numvout, vout);
							break;
						case 2 :		// textured
							zb_polygon3d_f(E3D_zbuffer, E3D_buffer, POLYTYPE_PTEX,
									scene->ptex[0], numvout, vout);
							break;
					}

// DEBUG ***
//blit(E3D_buffer, screen, 0, 0, 0, 0, 640, 480);
//rest(500);
//return 0;

					// DEBUG - display vert values
					if(key[KEY_V]) {
//						for(k=0; k<scene->pmodel[t1].poly[j].numverts; k++) {
//							fprintf(stderr, "Vert %d: addr %d   x %3.3f   y %3.3f   z %3.3f\n",
//								k, vout[k], vout[k]->x, vout[k]->y, vout[k]->z);
//						}
					}
//						fprintf(stderr, "Polys drawn: %d \n", polysdrawn);

				}	// end if polygon_z_normal > 0.0
			}	// next j (next poly)
		}	 // end if 2
	}	// end if 1
	}	// end if distance clip

}

// *******************************************************************
// E3D_Draw_ZBInst	- draw an instance, z-buffered (DEBUG!)
// *******************************************************************
void E3D_DrawZBInst(BITMAP *buffer, BITMAP *E3D_zbuffer, int instnum) {

}

// E3D_InstInRange - returns 1 if in range, 0 if not
int E3D_InstInRange(ThreeDeeInstance *pinst) {
	float distsqr, dx, dy, dz;
	dx = pinst->posn.x - E3D_scene.camera.posn.x;
	dy = pinst->posn.y - E3D_scene.camera.posn.y;
	dz = pinst->posn.z - E3D_scene.camera.posn.z;
   distsqr = dx*dx + dy*dy + dz*dz;
	if(distsqr < (E3D_scene.renderdist * E3D_scene.renderdist)) return 1;
	else return 0;
}

// *******************************************************************
// E3D_Draw_Scene	- draw a whole scene
// *******************************************************************
int E3D_DrawScene(int x_offset, int y_offset, int vblsync) {

	int i;
	Vector_f eyerotn;

	// set up this frame
	polysdrawn = 0; polysprocessed = 0;

	// clear background or set to picture
	if(E3D_scene.background == NULL) {
		i = makecol(E3D_scene.backcolour.r * 255, E3D_scene.backcolour.g * 255,
						E3D_scene.backcolour.b * 255);
		clear_to_color(E3D_buffer, i);
	}
	else blit(E3D_scene.background, E3D_buffer, 0, 0, 0, 0,
					E3D_scene.background->w, E3D_scene.background->h);

	clear(E3D_zbuffer);			// *** zbuffer specific

	// test zbuffer
//	rectfill(E3D_zbuffer, 50,50,100,100, 2000000000);

	// set up camera
	get_camera_matrix_f(&mfcamera,
		E3D_scene.camera.posn.x, E3D_scene.camera.posn.y, E3D_scene.camera.posn.z,
		E3D_scene.camera.dirn.x, E3D_scene.camera.dirn.y, E3D_scene.camera.dirn.z,
		E3D_scene.camera.up.x, E3D_scene.camera.up.y, E3D_scene.camera.up.z,
		FOV, 1.0);

	// render ground if loaded
	if(E3D_scene.tile.pdata != NULL)	E3D_RenderGround();

	// render landscape if loaded
	if(E3D_scene.plandmodel != NULL) E3D_DrawLandscape();

	// *** RENDER THE INSTANCES IN THE SCENE ***
	// (should do basic instance clipping here!)
	for(i=0; i < E3D_scene.numinsts; i++) {
		if(E3D_scene.pinst[i]->visible) {
			if(E3D_InstInRange(E3D_scene.pinst[i])) {
				E3D_DrawInst(E3D_scene.pinst[i]);
			}
		}
	}

	if(vblsync) vsync();
//	sprintf(mystring, "Getpixel %d  ", getpixel(E3D_zbuffer, 20, 20));
//	textout(E3D_buffer, font, mystring, 8, 460, makecol(255,255,255));
	sprintf(mystring, "polys %d / %d", polysdrawn, polysprocessed);
	textout(E3D_buffer, font, mystring, 8, 0, makecol(255,255,255));

	blit(E3D_buffer, screen, 0, 0, x_offset, y_offset, E3D_buffer->w, E3D_buffer->h);

	return 0;
}


// *******************************************************************
// E3D_SetGround	- set up ray-cast tiled ground (and activate it)
// *******************************************************************
// NB: Tiled ground can only with rotn.x = 0 and rotn.z = 0 
int E3D_SetGround(BITMAP *tilebmp) {
	int i, j, c;
	unsigned char *pd;

	E3D_scene.tile.w = tilebmp->w;
	E3D_scene.tile.h = tilebmp->h;

	// allocate memory for RGB tile data
	fprintf(stderr, "Allocating memory for tile data...\n");
	i = tilebmp->w * tilebmp->h * 4;
	E3D_scene.tile.pdata = malloc(i);

	// check if successful
	if(E3D_scene.tile.pdata == NULL) {
		fprintf(stderr, "Error: Malloc for tile data failed!\n");
		return 1;
	}
	else {
		fprintf(stderr, "Malloc for tile data OK!\n");
		// fill tile map RGB data
		fprintf(stderr, "Width: %d   Height: %d \n", tilebmp->w, tilebmp->h);
		pd = E3D_scene.tile.pdata;
		for(i=0; i<tilebmp->h; i++) {
			for(j=0; j<tilebmp->w; j++) {
				c = getpixel(tilebmp, j, i);
				*pd++ = (c >> 16) & 0xFF;			// R component
				*pd++ = (c >> 8) & 0xFF;			// G component
				*pd++ = (c >> 0) & 0xFF;			// B component
				pd++;
			}
		}
		fprintf(stderr, "Tile data loaded from bitmap.\n");
		return 0;
	} // end if
			
}

// *******************************************************************
// E3D_RenderGround	- render ray-cast tiled ground
// *******************************************************************
// NB: Tiled ground can only with rotn.x = 0 and rotn.z = 0
// NB: Uses only camera.posn.y and camera.rotn.y
int E3D_RenderGround(void) {
	int i, j, x, y, z, w, h, starty, c, tw, th;
	unsigned char r, g, b;
	float fr, fg, fb, dr, dg, db, fog_ratio;
	float d1, d4, fw, fh, fy, ry;
	float x1, x2, z1, z2, dx, dy, dz, dx1, dy1, dx2, dy2, alpha, dalpha, fov;
	float doffset[800];
	float startangle, endangle;
	unsigned char *pc;

	w = E3D_buffer->w;
	h = E3D_buffer->h;
	tw = E3D_scene.tile.w;
	th = E3D_scene.tile.h;

	fw = (float) w;


	fov = 48.0;

	// 1st, calculate starting and ending angles
	startangle = sin(E3D_scene.camera.dirn.x) + (fov/2);
	endangle = sin(E3D_scene.camera.dirn.x) - (fov/2);
	// calculate unit direction vector for starting angle
	dx1 = -sin(startangle / 57.3);
	dy1 = -cos(startangle / 57.3);
	// calculate unit direction vector for ending angle
	dx2 = -sin(endangle / 57.3);
	dy2 = -cos(endangle / 57.3);

	// calculate distance offsets for rays
	// and use for calculating start and end pts in texture
	starty = (h * 60) / 100;		//
	fh = E3D_scene.camera.posn.y;
	d1 = 2.0;								// distance to screen
	for(y=starty; y<h; y++) {
		fy = (float)(y - h/2) / 300.0;                			// scale this?
		d4 = ((fh-fy) / fy) * d1;		// dist from screen to pt.
		doffset[y] = d4 + d1;			// now have dist from eye to pt.
		//fprintf(stderr, "y %d   doffset %f\n", y, doffset[y]);
		// calculate start pt of this line in texture
		x1 = -E3D_scene.camera.posn.x + (dx1 * doffset[y]);
		z1 = -E3D_scene.camera.posn.z + (dy1 * doffset[y]);
		// calculate end pt of this line in texture
		x2 = -E3D_scene.camera.posn.x + (dx2 * doffset[y]);
		z2 = -E3D_scene.camera.posn.z + (dy2 * doffset[y]);
		// calculate increment
		dx = (x2 - x1) / fw; dz = (z2 - z1) / fw;

		// calculate fog effect (if on)
		fog_ratio = doffset[y] / E3D_scene.fogdistance;
		if(fog_ratio > 1.0) fog_ratio = 1.0;

		// draw horizontal line
		for(i=0; i<w; i++) {
			x = (int)x1 & 0x3F; z = (int)z1 & 0x3F;		// HACK to 64x64
			//x = abs(x); mod z = abs(z); 					// HACK!
			//fprintf(stderr, "%d  %d\n", x, z);
			pc = E3D_scene.tile.pdata + ((z * E3D_scene.tile.w) + x) * 4;
			r = *pc++; g = *pc++; b = *pc++;
			//fprintf(stderr, "%X  %X  %X\n", r, g, b);
			//interpolate fog colour
			fr = (float)r / 255.0;
			fg = (float)g / 255.0;
			fb = (float)b / 255.0;
			dr = (fr - E3D_scene.fogcolour.r) * fog_ratio;
			dg = (fg - E3D_scene.fogcolour.g) * fog_ratio;
			db = (fb - E3D_scene.fogcolour.b) * fog_ratio;
			c = makecol16((fr-dr) * 255, (fg-dg) * 255, (fb-db) * 255);
			putpixel(E3D_buffer, i, y, c);
			x1 += dx; z1 += dz;
		}
	}

}

// NB: make this fixed-pt version with no fog.
// speed-up idea: interpolate every 2nd pixel horizontally
int E3D_RenderGround2(void) {
	int i, j, x, y, z, w, h, starty, c;
	unsigned char r, g, b;
	unsigned char *pc;

	for(i=0; i<E3D_scene.tile.h; i++) {
		for(j=0; j<E3D_scene.tile.w; j++) {
			pc = E3D_scene.tile.pdata + ((i * E3D_scene.tile.w) + j) * 4;
			r = *pc++; g = *pc++; b = *pc++;
			//fprintf(stderr, "%X  %X  %X\n", r, g, b);
			c = makecol16(r, g, b);
			putpixel(E3D_buffer, j, i, c);
		}
	}

}
